/***************************************************
  Notebook for 240 x 320 panel with touchpanel

  2019' 09/30 Yuji Tanaka
 ****************************************************/

#include <Adafruit_GFX.h>    // Core graphics library
#include "FS.h"
#include "SD.h"
#include <SPI.h>
#include <Wire.h>      // this is needed even tho we aren't using it
#include <Adafruit_ILI9341.h>
//#include <Adafruit_STMPE610.h>
#include <XPT2046_Touchscreen.h>

// For STMPE610
// This is calibration data for the raw touch data to the screen coordinates
//#define TS_MINX 200
//#define TS_MINY 100
//#define TS_MAXX 3900
//#define TS_MAXY 3900

// The STMPE610 uses hardware SPI on the shield, and #8
//#define STMPE_CS 4
//Adafruit_STMPE610 ts = Adafruit_STMPE610(STMPE_CS);

// For XPT2046
#define CS_PIN 17
XPT2046_Touchscreen ts(CS_PIN);
#define TIRQ_PIN 16
#define TS_MINX 300
#define TS_MINY 300
#define TS_MAXX 3800
#define TS_MAXY 3800

#define TFT_DC  2
#define TFT_CS 15
#define TFT_RST 4
//#define TFT_RST -1

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_RST);

#define SD_CS 5
#define MAX_FILE 32
File root;
char fname[MAX_FILE][32];
char fnum;
int fcount;

// Size of the color selection boxes and the paintbrush size
#define BOXSIZE 30
int pen_radius;
int oldcolor, currentcolor;
int menu;  // 0-Color  1-Pen  2-File
int _width, _height;
uint16_t *buffer0;
uint16_t *buffer1;
uint16_t *buffer2;
uint16_t *buffer3;

void setup(void) {
  int i, j;
 // while (!Serial);     // used for leonardo debugging
 
  Serial.begin(115200);
  Serial.println(F("Touch Paint!"));
  
  tft.begin();
  tft.setRotation(0);
  tft.fillScreen(ILI9341_BLACK);
  _width=tft.width();
  _height=tft.height();
  buffer0 = (uint16_t *)malloc(_width*_height*2/4);
  buffer1 = (uint16_t *)malloc(_width*_height*2/4);
  buffer2 = (uint16_t *)malloc(_width*_height*2/4);
  buffer3 = (uint16_t *)malloc(_width*_height*2/4);
  for(j=0; j<_height/4; j++){
    for(i=0; i<_width; i++){
      buffer0[j*_width+i]=0;
      buffer1[j*_width+i]=0;
      buffer2[j*_width+i]=0;
      buffer3[j*_width+i]=0;
    }
  }

  if (!ts.begin()) {
    Serial.println("Couldn't start touchscreen controller");
    while (1);
  }
  ts.setRotation(0);
  Serial.println("Touchscreen started");
  
  Serial.print("Initializing SD card...");
  pinMode(SD_CS, PULLUP);
  pinMode(MISO, PULLUP);
  if (!SD.begin(SD_CS)) {
    Serial.println("failed!");
  }
  root = SD.open("/");
  printDirectory(root, 0, ".bmp");

  // make the color selection boxes
  show_boxes();
 
  // select the current color 'red'
  tft.drawRect(0, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
  currentcolor = ILI9341_RED;
  pen_radius = 2;
  menu = 0;
  fcount = 0;
}


void loop()
{
  int i, j;
  int sel;
  char str[16];
  
  if (ts.bufferEmpty() || !ts.touched()) {
    return;
  }

  TS_Point p = ts.getPoint();

  p.x = tft.width() - map(p.x, TS_MINX, TS_MAXX, 0, tft.width());
  p.y = tft.height() - map(p.y, TS_MINY, TS_MAXY, 0, tft.height());

  if (p.y < BOXSIZE) {
    oldcolor = currentcolor;

    if (menu==0){
      if (p.x < BOXSIZE) { 
        currentcolor = ILI9341_RED; 
        tft.drawRect(0, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
      } else if (p.x < BOXSIZE*2) {
        currentcolor = ILI9341_YELLOW;
        tft.drawRect(BOXSIZE, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
      } else if (p.x < BOXSIZE*3) {
        currentcolor = ILI9341_GREEN;
        tft.drawRect(BOXSIZE*2, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
      } else if (p.x < BOXSIZE*4) {
        currentcolor = ILI9341_CYAN;
        tft.drawRect(BOXSIZE*3, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
      } else if (p.x < BOXSIZE*5) {
        currentcolor = ILI9341_BLUE;
        tft.drawRect(BOXSIZE*4, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
      } else if (p.x < BOXSIZE*6) {
        currentcolor = ILI9341_MAGENTA;
        tft.drawRect(BOXSIZE*5, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
      } else if (p.x < BOXSIZE*7) {
        currentcolor = ILI9341_WHITE;
        tft.drawRect(BOXSIZE*6, 0, BOXSIZE, BOXSIZE, ILI9341_RED);
      } else if (p.x < BOXSIZE*8) {
        menu = 1;
        show_boxes();
      }

      if (oldcolor != currentcolor) {
        if (oldcolor == ILI9341_RED) 
          tft.fillRect(0, 0, BOXSIZE, BOXSIZE, ILI9341_RED);
        if (oldcolor == ILI9341_YELLOW) 
           tft.fillRect(BOXSIZE, 0, BOXSIZE, BOXSIZE, ILI9341_YELLOW);
         if (oldcolor == ILI9341_GREEN) 
           tft.fillRect(BOXSIZE*2, 0, BOXSIZE, BOXSIZE, ILI9341_GREEN);
         if (oldcolor == ILI9341_CYAN) 
           tft.fillRect(BOXSIZE*3, 0, BOXSIZE, BOXSIZE, ILI9341_CYAN);
         if (oldcolor == ILI9341_BLUE) 
           tft.fillRect(BOXSIZE*4, 0, BOXSIZE, BOXSIZE, ILI9341_BLUE);
         if (oldcolor == ILI9341_MAGENTA) 
           tft.fillRect(BOXSIZE*5, 0, BOXSIZE, BOXSIZE, ILI9341_MAGENTA);
         if (oldcolor == ILI9341_WHITE) 
           tft.fillRect(BOXSIZE*6, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
      }
    }else if (menu==1){
      if (p.x < BOXSIZE)
        pen_radius = 1;
      else if (p.x < BOXSIZE*2)
        pen_radius = 2;
      else if (p.x < BOXSIZE*3)
        pen_radius = 3;
      else if (p.x < BOXSIZE*4)
        pen_radius = 4;
      else if (p.x < BOXSIZE*5)
        pen_radius = 5;
      else if (p.x < BOXSIZE*6)
        pen_radius = 6;
      else if (p.x < BOXSIZE*7)
        pen_radius = 7;
      else if (p.x < BOXSIZE*8)
        menu = 2;
      show_boxes();
      tft.drawRect(BOXSIZE*(pen_radius-1), 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    }else if (menu==2){
      // Clear
      if (p.x < BOXSIZE){
        tft.fillScreen(ILI9341_BLACK);
        for(j=0; j<_height/4; j++){
          for(i=0; i<_width; i++){
            buffer0[j*_width+i]=0;
            buffer1[j*_width+i]=0;
            buffer2[j*_width+i]=0;
            buffer3[j*_width+i]=0;
          }
        }
      } else if (p.x < BOXSIZE*2) {
        tft.fillRect(0, 0, BOXSIZE*8, BOXSIZE, ILI9341_BLACK);
        tft.setTextSize(2);
        tft.setTextColor(ILI9341_WHITE);
        tft.setCursor(0, 4);
        tft.print("** Now Saving ! **");
        sprintf(str, "/note%d.bmp", fcount % 10);
        fcount++;
        bmpSave(str, _width, _height-BOXSIZE);
      } else if (p.x < BOXSIZE*3) {
        sel=select_file();
        if (sel>=0){
          tft.fillScreen(ILI9341_BLACK);
          for(j=0; j<_height/4; j++){
            for(i=0; i<_width; i++){
              buffer0[j*_width+i]=0;
              buffer1[j*_width+i]=0;
              buffer2[j*_width+i]=0;
              buffer3[j*_width+i]=0;
            }
          }
          tft.setTextSize(2);
          tft.setTextColor(ILI9341_WHITE);
          tft.setCursor(0, 4);
          tft.print("** Now Loading ! **");
          bmpDraw(fname[sel], 0, BOXSIZE);
        }
      } else if (p.x < BOXSIZE*8) {
        menu = 0;
      }
      show_boxes();
    }
    delay(200);
  }

  if (((p.y-pen_radius) > BOXSIZE) && ((p.y+pen_radius) < tft.height())) {
    tft.fillCircle(p.x, p.y, pen_radius, currentcolor);
    buf_fillCircle(p.x, p.y, pen_radius, currentcolor);
  }
}

void show_boxes()
{
  int i;
  if (menu==0){
    tft.fillRect(0, 0, BOXSIZE*8, BOXSIZE, ILI9341_BLACK);
    tft.fillRect(0, 0, BOXSIZE, BOXSIZE, ILI9341_RED);
    tft.fillRect(BOXSIZE, 0, BOXSIZE, BOXSIZE, ILI9341_YELLOW);
    tft.fillRect(BOXSIZE*2, 0, BOXSIZE, BOXSIZE, ILI9341_GREEN);
    tft.fillRect(BOXSIZE*3, 0, BOXSIZE, BOXSIZE, ILI9341_CYAN);
    tft.fillRect(BOXSIZE*4, 0, BOXSIZE, BOXSIZE, ILI9341_BLUE);
    tft.fillRect(BOXSIZE*5, 0, BOXSIZE, BOXSIZE, ILI9341_MAGENTA);
    tft.fillRect(BOXSIZE*6, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    tft.fillRect(BOXSIZE*7, 0, BOXSIZE, BOXSIZE, ILI9341_BLACK);
    if (currentcolor==ILI9341_RED)
      tft.drawRect(0, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    if (currentcolor==ILI9341_YELLOW)
      tft.drawRect(BOXSIZE, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    if (currentcolor==ILI9341_GREEN)
      tft.drawRect(BOXSIZE*2, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    if (currentcolor==ILI9341_CYAN)
      tft.drawRect(BOXSIZE*3, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    if (currentcolor==ILI9341_BLUE)
      tft.drawRect(BOXSIZE*4, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    if (currentcolor==ILI9341_MAGENTA)
      tft.drawRect(BOXSIZE*5, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    if (currentcolor==ILI9341_WHITE)
      tft.drawRect(BOXSIZE*6, 0, BOXSIZE, BOXSIZE, ILI9341_RED);
    tft.drawRect(BOXSIZE*7, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    tft.setTextSize(1);
    tft.setTextColor(ILI9341_WHITE);
    tft.setCursor(BOXSIZE*7, 12);
    tft.print("Color");
  }else if (menu==1){
    tft.fillRect(0, 0, BOXSIZE*8, BOXSIZE, ILI9341_BLACK);
    for(i=0;i<7;i++){
      if (pen_radius==i+1)
        tft.drawRect(BOXSIZE*i, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
      else
        tft.drawRect(BOXSIZE*i, 0, BOXSIZE, BOXSIZE, ILI9341_DARKGREY);
      tft.fillCircle(i*BOXSIZE+BOXSIZE/2, BOXSIZE/2, i+1, currentcolor);
    }
    tft.drawRect(BOXSIZE*7, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);
    tft.setTextSize(1);
    tft.setTextColor(ILI9341_WHITE);
    tft.setCursor(BOXSIZE*7+6, 12);
    tft.print("Pen");
  }else if (menu==2){
    tft.fillRect(0, 0, BOXSIZE*8, BOXSIZE, ILI9341_BLACK);
    for(i=0;i<8;i++)
      tft.drawRect(BOXSIZE*i, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);

    tft.setTextSize(1);
    tft.setTextColor(ILI9341_WHITE);

    tft.setCursor(0, 12);
    tft.print("Clear");
    tft.setCursor(BOXSIZE+3, 12);
    tft.print("Save");
    tft.setCursor(BOXSIZE*2+3, 12);
    tft.print("Load");
    tft.setCursor(BOXSIZE*7+3, 12);
    tft.print("File");
  }
}

int select_file()
{
  int n, old_n;
  
  tft.fillRect(0, 0, BOXSIZE*8, BOXSIZE, ILI9341_BLACK);
  for(n=4; n<8; n++)
    tft.drawRect(BOXSIZE*n, 0, BOXSIZE, BOXSIZE, ILI9341_WHITE);

  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(BOXSIZE*4+12, 12);
  tft.print("<");
  tft.setCursor(BOXSIZE*5+12, 12);
  tft.print(">");
  tft.setCursor(BOXSIZE*6+3, 12);
  tft.print("Skip");
  tft.setCursor(BOXSIZE*7+9, 12);
  tft.print("OK");
  n=0;
  old_n=-1;

  for(;;){
    if (n!=old_n){
      tft.fillRect(0, 0, BOXSIZE*4, BOXSIZE, ILI9341_BLACK);
      tft.setCursor(0, 12);
      tft.print(fname[n]);
      old_n=n;
    }
    
    if (ts.bufferEmpty() || !ts.touched()) {
      delay(100);
      continue;
    }

    TS_Point p = ts.getPoint();

    p.x = tft.width() - map(p.x, TS_MINX, TS_MAXX, 0, tft.width());
    p.y = tft.height() - map(p.y, TS_MINY, TS_MAXY, 0, tft.height());

    if (p.y < BOXSIZE) {
      if ((p.x/BOXSIZE)==4){
        n--;
        if (n<0)
          n=fnum-1;
      }
      if ((p.x/BOXSIZE)==5){
        n++;
        if (n>=fnum)
          n=0;
      }
      if ((p.x/BOXSIZE)==6)
        return(-1);
      if ((p.x/BOXSIZE)==7)
        return(n);
    }
    delay(100);
  }
}

void printDirectory(File dir, int numTabs, char *typestr) {
  fnum=0;
  while (true) {

    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      break;
    }
    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }
    Serial.print(entry.name());
    if (fnum<MAX_FILE && stristr(entry.name(), typestr)!=NULL){
      strcpy(fname[fnum++], entry.name());
    }else if(fnum>=MAX_FILE){
      entry.close();
      return;
    }
    
    if (entry.isDirectory()) {
      Serial.println("/");
      printDirectory(entry, numTabs + 1, typestr);
    } else {
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
    }
    entry.close();
  }
}

// This function opens a Windows Bitmap (BMP) file and
// displays it at the given coordinates.  It's sped up
// by reading many pixels worth of data at a time
// (rather than pixel by pixel).  Increasing the buffer
// size takes more of the Arduino's precious RAM but
// makes loading a little faster.  20 pixels seems a
// good balance.

#define BUFFPIXEL 20

void bmpDraw(char *filename, int16_t x, int16_t y) {

  File     bmpFile;
  int      bmpWidth, bmpHeight;   // W+H in pixels
  uint8_t  bmpDepth;              // Bit depth (currently must be 24)
  uint32_t bmpImageoffset;        // Start of image data in file
  uint32_t rowSize;               // Not always = bmpWidth; may have padding
  uint8_t  sdbuffer[3*BUFFPIXEL]; // pixel buffer (R+G+B per pixel)
  uint8_t  buffidx = sizeof(sdbuffer); // Current position in sdbuffer
  boolean  goodBmp = false;       // Set to true on valid header parse
  boolean  flip    = true;        // BMP is stored bottom-to-top
  int      w, h, row, col, x2, y2, bx1, by1;
  uint8_t  r, g, b;
  uint32_t pos = 0, startTime = millis();

  if((x >= tft.width()) || (y >= tft.height())) return;

  Serial.println();
  Serial.print(F("Loading image '"));
  Serial.print(filename);
  Serial.println('\'');

  // Open requested file on SD card
  if ((bmpFile = SD.open(filename)) == NULL) {
    Serial.print(F("File not found"));
    return;
  }

  // Parse BMP header
  if(read16(bmpFile) == 0x4D42) { // BMP signature
    Serial.print(F("File size: ")); Serial.println(read32(bmpFile));
    (void)read32(bmpFile); // Read & ignore creator bytes
    bmpImageoffset = read32(bmpFile); // Start of image data
    Serial.print(F("Image Offset: ")); Serial.println(bmpImageoffset, DEC);
    // Read DIB header
    Serial.print(F("Header size: ")); Serial.println(read32(bmpFile));
    bmpWidth  = read32(bmpFile);
    bmpHeight = read32(bmpFile);
    if(read16(bmpFile) == 1) { // # planes -- must be '1'
      bmpDepth = read16(bmpFile); // bits per pixel
      Serial.print(F("Bit Depth: ")); Serial.println(bmpDepth);
      if((bmpDepth == 24) && (read32(bmpFile) == 0)) { // 0 = uncompressed

        goodBmp = true; // Supported BMP format -- proceed!
        Serial.print(F("Image size: "));
        Serial.print(bmpWidth);
        Serial.print('x');
        Serial.println(bmpHeight);

        // BMP rows are padded (if needed) to 4-byte boundary
        rowSize = (bmpWidth * 3 + 3) & ~3;

        // If bmpHeight is negative, image is in top-down order.
        // This is not canon but has been observed in the wild.
        if(bmpHeight < 0) {
          bmpHeight = -bmpHeight;
          flip      = false;
        }

        // Crop area to be loaded
        x2 = x + bmpWidth  - 1; // Lower-right corner
        y2 = y + bmpHeight - 1;
        if((x2 >= 0) && (y2 >= 0)) { // On screen?
          w = bmpWidth; // Width/height of section to load/display
          h = bmpHeight;
          bx1 = by1 = 0; // UL coordinate in BMP file
          if(x < 0) { // Clip left
            bx1 = -x;
            x   = 0;
            w   = x2 + 1;
          }
          if(y < 0) { // Clip top
            by1 = -y;
            y   = 0;
            h   = y2 + 1;
          }
          if(x2 >= tft.width())  w = tft.width()  - x; // Clip right
          if(y2 >= tft.height()) h = tft.height() - y; // Clip bottom
  
          // Set TFT address window to clipped image bounds
          tft.startWrite(); // Requires start/end transaction now
//          tft.setAddrWindow(x, y, w, h);
  
          for (row=0; row<h; row++) { // For each scanline...
  
            // Seek to start of scan line.  It might seem labor-
            // intensive to be doing this on every line, but this
            // method covers a lot of gritty details like cropping
            // and scanline padding.  Also, the seek only takes
            // place if the file position actually needs to change
            // (avoids a lot of cluster math in SD library).
            if(flip) // Bitmap is stored bottom-to-top order (normal BMP)
              pos = bmpImageoffset + (bmpHeight - 1 - (row + by1)) * rowSize;
            else     // Bitmap is stored top-to-bottom
              pos = bmpImageoffset + (row + by1) * rowSize;
            pos += bx1 * 3; // Factor in starting column (bx1)
            if(bmpFile.position() != pos) { // Need seek?
              tft.endWrite(); // End TFT transaction
              bmpFile.seek(pos);
              buffidx = sizeof(sdbuffer); // Force buffer reload
              tft.startWrite(); // Start new TFT transaction
            }
            for (col=0; col<w; col++) { // For each pixel...
              // Time to read more pixel data?
              if (buffidx >= sizeof(sdbuffer)) { // Indeed
                tft.endWrite(); // End TFT transaction
                bmpFile.read(sdbuffer, sizeof(sdbuffer));
                buffidx = 0; // Set index to beginning
                tft.startWrite(); // Start new TFT transaction
              }
              // Convert pixel from BMP to TFT format, push to display
              b = sdbuffer[buffidx++];
              g = sdbuffer[buffidx++];
              r = sdbuffer[buffidx++];
//              tft.writePixel(tft.color565(r,g,b));
              tft.writePixel(col, y+row, tft.color565(r,g,b));
              buf_writePixel(col, y+row, tft.color565(r,g,b));
            } // end pixel
          } // end scanline
          tft.endWrite(); // End last TFT transaction
        } // end onscreen
        Serial.print(F("Loaded in "));
        Serial.print(millis() - startTime);
        Serial.println(" ms");
      } // end goodBmp
    }
  }

  bmpFile.close();
  if(!goodBmp) Serial.println(F("BMP format not recognized."));
}

// These read 16- and 32-bit types from the SD card file.
// BMP data is stored little-endian, Arduino is little-endian too.
// May need to reverse subscript order if porting elsewhere.

uint16_t read16(File &f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File &f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}

char *stristr(const char *org, char *str)
{
  char *tmp, *tmp0, *tmp1;
  tmp=(char *)org;
  for(;*tmp!='\0';tmp++){
    tmp0=tmp;
    tmp1=str;
    for(;*tmp1!='\0';tmp1++){
      if (*tmp0==*tmp1){
        tmp0++;
        continue;
      }else if ('A'<=*tmp1 && *tmp1<='Z' && *tmp0==(*tmp1-'A'+'a')){
        tmp0++;
        continue;
      }else if ('a'<=*tmp1 && *tmp1<='z' && *tmp0==(*tmp1-'a'+'A')){
        tmp0++;
        continue;
      }else
        break;
    }
    if (*tmp1=='\0')
      return(tmp);
  }
  return(NULL);
}

void bmpSave(char *filename, int16_t x, int16_t y)
{
  File     bmpFile;
  int i, j;
  uint8_t buf[4];
  uint8_t bmp_header[54]=
   {0x42,0x4d,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x36,0x00,0x00,0x00,0x28,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x18,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x80,0x00,0x00,0x00,0x80,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00};
  int skip_cnt;
  int calc;
  
  Serial.println();
  Serial.print(F("Saving image '"));
  Serial.print(filename);
  Serial.println('\'');

  // Open requested file on SD card
  if ((bmpFile = SD.open(filename, FILE_WRITE)) == NULL) {
    Serial.print(F("Cannot write File !"));
    return;
  }
  
  skip_cnt=4-(x*3) % 4;
  if(skip_cnt==4)
    skip_cnt=0;
  calc=(x*3+skip_cnt)*y+0x36;
  bmp_header[2]=calc % 256;
  bmp_header[3]=(calc/256) % 256;
  bmp_header[4]=(calc/256/256) % 256;
  bmp_header[5]=(calc/256/256/256) % 256;
  bmp_header[18]=x % 256;
  bmp_header[19]=(x/256) % 256;
  bmp_header[20]=(x/256/256) % 256;
  bmp_header[21]=(x/256/256/256) % 256;
  bmp_header[22]=y % 256;
  bmp_header[23]=(y/256) % 256;
  bmp_header[24]=(y/256/256) % 256;
  bmp_header[25]=(y/256/256/256) % 256;
  calc=(x*3+skip_cnt)*y;
  bmp_header[34]=calc % 256;
  bmp_header[35]=(calc/256) % 256;
  bmp_header[36]=(calc/256/256) % 256;
  bmp_header[37]=(calc/256/256/256) % 256;
  bmpFile.write(bmp_header, 54);
  for(i=_height-1;i>=(_height-y);i--){
    for(j=0;j<x;j++){
      if (i<_height/4)
        calc=buffer0[i*_width+j];
      else if (i<_height*2/4)
        calc=buffer1[(i-_height/4)*_width+j];
      else if (i<_height*3/4)
        calc=buffer2[(i-_height*2/4)*_width+j];
      else
        calc=buffer3[(i-_height*3/4)*_width+j];
      buf[0]=(calc & 0x001F) << 3;
      buf[1]=(calc & 0x07E0) >> 3;
      buf[2]=(calc & 0xF800) >> 8;
      bmpFile.write(buf, 3);
    }
    if (skip_cnt>0){
      for(j=0; j<4; j++)
        buf[j]=0;
      bmpFile.write(buf, skip_cnt);
    }  
  }

  bmpFile.close();
}

void buf_writePixel(int16_t x, int16_t y, uint16_t color)
{
  uint16_t i;
  uint16_t m, n;

  if (x<0 || x>=_width || y<0 || y>=_height)
    return;

  m=y * 4 / _height;
  n=y - m * _height / 4;

  i=(n * _width) + x;
  if (m==0){
    buffer0[i]=color;
  }else if (m==1){
    buffer1[i]=color;
  }else if (m==2){
    buffer2[i]=color;
  }else{
    buffer3[i]=color;
  }
}

void buf_writeFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color)
{
  uint16_t i, j;
  uint16_t m, n;

  if (x<0 || x>=_width || y>=_height)
    return;
  if (y+h>=_height)
    h=_height-y;

  for(j=y; j<y+h; j++){
    m=j * 4 / _height;
    n=j - m * _height / 4;

    i=(n * _width) + x;
    if (m==0){
      buffer0[i]=color;
    }else if (m==1){
      buffer1[i]=color;
    }else if (m==2){
      buffer2[i]=color;
    }else{
      buffer3[i]=color;
    }
  }
}

void buf_fillCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color)
{
    buf_writeFastVLine(x0, y0-r, 2*r+1, color);
    buf_fillCircleHelper(x0, y0, r, 3, 0, color);
}

void buf_fillCircleHelper(int16_t x0, int16_t y0, int16_t r,
  uint8_t corners, int16_t delta, uint16_t color) {

    int16_t f     = 1 - r;
    int16_t ddF_x = 1;
    int16_t ddF_y = -2 * r;
    int16_t x     = 0;
    int16_t y     = r;
    int16_t px    = x;
    int16_t py    = y;

    delta++; // Avoid some +1's in the loop

    while(x < y) {
        if (f >= 0) {
            y--;
            ddF_y += 2;
            f     += ddF_y;
        }
        x++;
        ddF_x += 2;
        f     += ddF_x;
        if(x < (y + 1)) {
            if(corners & 1) buf_writeFastVLine(x0+x, y0-y, 2*y+delta, color);
            if(corners & 2) buf_writeFastVLine(x0-x, y0-y, 2*y+delta, color);
        }
        if(y != py) {
            if(corners & 1) buf_writeFastVLine(x0+py, y0-px, 2*px+delta, color);
            if(corners & 2) buf_writeFastVLine(x0-py, y0-px, 2*px+delta, color);
            py = y;
        }
        px = x;
    }
}
